@@ -34,13 +34,14 @@
 #include "gok-gconf.h"
 #include "main.h"
 
+#ifdef HAVE_XINPUT
+static  XDeviceInfo  *devices = NULL;
+#endif
+
 GSList *
 gok_input_get_device_list (void)
 {
   GSList       *pList = NULL;
-#ifdef HAVE_XINPUT
-  XDeviceInfo  *devices;
-#endif
   int		i;
   int		num_devices;
   Display      *display;
@@ -49,17 +50,21 @@ gok_input_get_device_list (void)
 #ifdef HAVE_XINPUT
   if (pMainWindow && pMainWindow->window) {
 	  display = GDK_WINDOW_XDISPLAY (pMainWindow->window); 
+	  if (devices) {
+		XFreeDeviceList (devices);
+		devices = NULL;
+	  }
+ 
 	  devices = XListInputDevices(display, &num_devices);
 	  for (i = 0; i < num_devices; i++) {
 		  GokInput *pInput;
 		  if (devices[i].use == IsXExtensionDevice) {
 			  pInput = g_new0 (GokInput, 1);
-			  pInput->name = devices[i].name;
+			  pInput->name = g_strdup (devices[i].name);
 			  pInput->info = &devices[i];
 			  pList = g_slist_prepend (pList, pInput);
 		  }
 	  }
-	  /* TODO: do we need to free the XDeviceInfo list? */
   }
 #endif
   return pList;
@@ -68,32 +73,30 @@ gok_input_get_device_list (void)
 #ifdef HAVE_XINPUT
 static XDeviceInfo*
 find_device_info(Display	*display,
-		 char		*name)
+		 char		*name,
+		 gboolean       extended_only)
 {
-	XDeviceInfo	*devices;
 	int		i;
 	int		num_devices;
-	int		len = strlen(name);
 	
+	if (devices) {
+		XFreeDeviceList (devices);
+		devices = NULL;
+	}
+
 	devices = XListInputDevices(display, &num_devices);
 	
 	for(i = 0; i < num_devices; i++) {
+		if (extended_only)
+			if (devices[i].use != IsXExtensionDevice)
+				continue;
+
 		if (!strcmp(devices[i].name, name))
 			return &devices[i];
 	}
 	return NULL;
 }
 
-
-static XDeviceInfo *
-gok_get_input_device (Display *display)
-{
-	char *input_device_name;
-	gok_gconf_get_string (gconf_client_get_default (), 
-			      GOK_GCONF_INPUT_DEVICE, 
-			      &input_device_name);
-	return find_device_info (display, input_device_name);
-}
 #endif
 
 gboolean
@@ -150,7 +153,7 @@ gok_input_init (GokInput *input, GdkFilterFunc filter_func)
 		return FALSE;
 	}
 	
-	gdk_window_add_filter (root,
+	gdk_window_add_filter (NULL,
 			       filter_func,
 			       NULL);
 	return TRUE;
@@ -184,8 +187,8 @@ gok_input_get_extension_device_name (void)
 void
 gok_input_free (GokInput *pInput)
 {
-	/* FIXME */
-	;
+	g_free (pInput->name);
+	g_free (pInput);
 }
 
 GokInput * 
@@ -198,14 +201,16 @@ gok_input_find_by_name (char *name, gboolean extended_only)
   
   if (window && window->window) 
     {
-      device = find_device_info (GDK_WINDOW_XDISPLAY (window->window), name);
+      device = find_device_info (GDK_WINDOW_XDISPLAY (window->window), name, extended_only);
     }
-  if (device && (device->use == IsXExtensionDevice) || !extended_only)
+  if (device)
     {
       input = g_new0 (GokInput, 1); 
       input->info = device;
-      input->name = g_strndup (name, 128);
+      input->name = g_strdup (device->name);
     }
+#else
+  g_message ("Warning, XInput extension not present, cannot open input device.");
 #endif
   return input;
 }
@@ -223,13 +228,18 @@ gok_input_open (GokInput *input)
       return TRUE;
     }
 #endif
+  g_warning ("could not open device %s", input->name);
   return FALSE;
 }
 
 GokInput * gok_input_get_current (void)
 {
   char *input_device_name;
-  GokInput *input;
+  static GokInput *input = NULL;
+
+  if (input)
+    gok_input_free (input);
+
   gok_gconf_get_string (gconf_client_get_default (), 
 			GOK_GCONF_INPUT_DEVICE, 
 			&input_device_name);
